package experiment;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.Set;

import org.graphstream.graph.Graph;
import org.graphstream.graph.implementations.SingleGraph;
import org.graphstream.stream.file.FileSinkDOT;

public class Main {
	
	private static List<Agent> agents;
	private static List<Agent> children;
	private static Map<String, double[]> history; // A map from an agent pair (at the format "id-id"; non-active children called "cid", e.g. c19) to an array at the format [dialogs, successes]
	
	private static Random random = new Random();
	private static int numberOfAgents;
	private static int numberOfGenerations;
	private static double epsilon;
	private static int numberOfConcepts;
	
	public static void main(String[] args) {
		numberOfGenerations = 100;
		numberOfConcepts = 2;
		int[] noAgents = {10};
		for (int nAgents : noAgents) {
			//epsilon = 0.0001; // 10^-4
			//epsilon = 0.00001; // 10^-5
			epsilon = Math.sqrt(0.00001 / nAgents);
			numberOfAgents = nAgents;
			double[][] results = new double[10][numberOfGenerations]; // ke tilbake til 10 hvis snitt + ke til 3*nOG hvis tre ord
			//double[][] results = new double[10][100];
			for (int i = 0; i < 1; i++) { // ke
				results[i] = game();
				//System.out.println("GAME, gen1: " + results[i][0]);
			}
			//System.out.println(nAgents + " agents:");
			for (int i = 0; i < numberOfGenerations; i++) { // endre til 3*nOG hvis tre ord
			//for (int i = 0; i < 100; i++) {
				double sum = 0;
				for (int j = 0; j < 9; j++) { // ke
					sum += results[j][i];
				}
				//System.out.println(sum / 9.0); // ke
			}
			/*for (Agent agent : agents) {
				System.out.print(agent);
				for (int i = 0; i < agent.getNumberOfConcepts(); i++) {
					System.out.print(", " + i + ":");
					for (String word : agent.getVocabulary(i))
						System.out.print(" " + word);
				}
				System.out.println();
			}*/
			
			/*FileSinkDOT fs = new FileSinkDOT();
			try {
				fs.writeAll(graph, "");
			} catch (IOException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}*/
		}
	}

	public static double[] game() {
		List<Integer> displayGraphs = Arrays.asList(10, 50, 100, 200);
		
		double[] results = new double[numberOfGenerations]; // endre til 3* hvis 3 ord
		//double[] results = new double[100];
		
		agents = new ArrayList<>();
		history = new HashMap<>();
		for (int i = 0; i < numberOfAgents; i++) {
			Agent newAgent = new Agent(numberOfConcepts, numberOfAgents, i);
			agents.add(newAgent);
			for (int j = 0; j < i; j++) {
				String agentPair = i + "-" + j;
				double[] historyPair = new double[] {0, 0};
				history.put(agentPair, historyPair);
			}
		}
		
		/*for (Agent[] pair : history.keySet()) {
			System.out.println(pair[0].id + ":" + pair[1].id);
		}*/
		
		children = new ArrayList<>();
		
		for (int i = 0; i < numberOfGenerations; i++) {
			//System.out.println("gen " + i);
			//if (i % 100 == 0)
				//System.out.println("gen " + i);
			int numberOfSuccesses = 0;
			int numberOfDialogs = 0;
			while (true) {
				numberOfDialogs++;
				//System.out.println(children.size());
				if (performDialog())
					numberOfSuccesses++;
				if (children.size() == agents.size())
					break;
			}
			Map<String, Integer> totalVocabulary0 = new HashMap<>();
			for (int j = 0; j < numberOfAgents; j++) {
				//System.out.print(agents.get(j).getVocabularySize());
				Set<String> agentVocabulary = agents.get(j).getVocabulary(0);
				for (String word : agentVocabulary) {
					//System.out.println("ag" + j + ": " + word);
					if (totalVocabulary0.containsKey(word))
						totalVocabulary0.put(word, totalVocabulary0.get(word) + 1);
					else
						totalVocabulary0.put(word, 1);
				}
			}
			/*Map<String, Integer> totalVocabulary1 = new HashMap<>();
			for (int j = 0; j < numberOfAgents; j++) {
				//System.out.print(agents.get(j).getVocabularySize());
				Set<String> agentVocabulary = agents.get(j).getVocabulary(1);
				for (String word : agentVocabulary) {
					//System.out.println("ag" + j + ": " + word);
					if (totalVocabulary1.containsKey(word))
						totalVocabulary1.put(word, totalVocabulary1.get(word) + 1);
					else
						totalVocabulary1.put(word, 1);
				}
			}*/
			
			agents.clear();
			agents.addAll(children);
			children.clear();
			generationalHistoryCleanup();
			
			/*if (displayGraphs.contains(i + 1)) {
				Graph graph = new SingleGraph("Experiments");
				for (int j = 0; j < agents.size(); j++) {
					graph.addNode(Integer.toString(j));
				}
				for (int j = 0; j < agents.size(); j++) {
					for (int k = 0; k < j; k++) {
						if (getSuccessRate(j, k) > 0.5)
							graph.addEdge(Integer.toString(j) + "-" + Integer.toString(k), Integer.toString(j), Integer.toString(k));
					}
				}
				graph.display();
			}*/
			
			/*if (totalVocabulary.size() == 1) {
				for (int j = i; j < numberOfGenerations; j++) {
					results[j] = 1;
					//results[j] = 1 / numberOfAgents;
				}
				break;
			}*/
			/*List<Integer> languages = new ArrayList<>(totalVocabulary.values());
			Collections.sort(languages);
			Collections.reverse(languages);
			results[3*i] = 1.0 * languages.get(0) / numberOfAgents;
			if (languages.size() > 1)
				results[3*i+1] = 1.0 * languages.get(1) / numberOfAgents;
			else
				results[3*i+1] = 0;
			if (languages.size() > 2)
			results[3*i+2] = 1.0 * languages.get(2) / numberOfAgents;
			else
				results[3*i+2] = 0;*/ //PRINT THREE MOST USED WORDS (remember to change instances of results to length 3*i)
			//results[i] = 1.0 * Collections.max(totalVocabulary0.values()) / numberOfAgents; // PRINT SHARE OF MOST COMMON LANGUAGE
			//results[i] = totalVocabulary0.size(); // PRINT NUMBER OF LANGUAGES
			//results[i] = 1.0 * totalVocabulary0.size() / numberOfAgents; // PRINT RELATIVE NUMBER OF LANGUAGES
			//results[i] = 1.0 * numberOfSuccesses / numberOfDialogs; // PRINT SUCCESS RATE
			//System.out.println("GENERATION " + i);
			//System.out.println("dialogs: " + numberOfDialogs);
			//System.out.println("successes: " + numberOfSuccesses);
			//System.out.println("vocab: " + totalVocabulary0.size());
			//System.out.println("vocab 2: " + totalVocabulary1.size());
		}
		/*double[] percentiles = new double[100];
		for (int i = 0; i < agents.size(); i++) {
			for (int j = 0; j < i; j++) {
				int percentile = (int) (getSuccessRate(i, j) * 100);
				percentiles[percentile]++;
			}
		}
		return percentiles;*/
		//System.out.println(numberOfSuccesses);
		//for (int i = 0; i < pairDialogs.length; i++) {
			//for (int j = 0; j < pairDialogs[i].length; j++) {
				//if (pairDialogs[i][j] > 0) {}
					//System.out.print(pairSuccesses[i][j] / pairDialogs [i][j]);
				//System.out.print(";");
			//}
			//System.out.println();
		//}
		return results; //<------------------------------------------------------------------- HUSK
	}
	
	private static boolean performDialog() {
		int speakerId = random.nextInt(numberOfAgents);
		Agent speaker = agents.get(speakerId);
		int hearerId = selectHearerId(speakerId);
		Agent hearer = agents.get(hearerId);
		//System.out.println("speaker " + speakerId + ", hearer " + hearerId);
		incrementDialogs(speakerId, hearerId);
		boolean success = dialog(speaker, hearer);
		if (success) {
			incrementSuccesses(speakerId, hearerId);
			Agent child = EA.reproduction(speaker, hearer, agents.size());
			children.add(child);
			addHistoryForChild(child, children.size() - 1, speakerId, hearerId);
		}
		
		return success;
	}

	private static int selectHearerId(int speakerId) {
		//System.out.println("DIALOG");
		//System.out.println("speaker: " + speakerId);
		double totalWeights = 0;
		for (int i = 0; i < agents.size(); i++) {
			if (i == speakerId)
				continue;
			totalWeights += epsilon;
			totalWeights += getSuccessRate(speakerId, i, false);
		}
		double randomNumber = totalWeights * random.nextDouble();
		double temp = 0;
		for (int i = 0; i < agents.size(); i++) {
			if (i == speakerId)
				continue;
			temp += epsilon;
			temp += getSuccessRate(speakerId, i, false);
			if (temp > randomNumber) {
				//System.out.println("hearer: " + i);
				getSuccessRate(speakerId, i, true);
				return i;
			}
		}
		return -1;
	}
	
	private static boolean dialog(Agent speaker, Agent hearer) {
		int concept = random.nextInt(numberOfConcepts);
		//System.out.println("concept " + concept);
		String word = speaker.speak(concept);
		boolean success = hearer.listen(concept, word);
		//System.out.println("word: " + word);
		if (success) {
			speaker.retainOnly(concept, word);
			hearer.retainOnly(concept, word);
		}
		else
			hearer.addWord(concept, word);
		//System.out.print("DIALOG: " + word + " - ");
		//System.out.println(success ? "SUCCESS" : "FAILIURE");
		//System.out.println(success ? 1 : 0);
//		if (success)
//			System.out.println("success");
		//System.out.println("success: " + success);
		return success;
	}
	
	private static double getSuccessRate(int agent1, int agent2, boolean print) {
		/*System.out.println("size " + history.size());
		for (String pair : history.keySet()) {
			System.out.println(pair);
		}
		/*for (int[] pair : history.values()) {
			System.out.println(pair[0] + ":" + pair[1]);
		}*/
		
		
		//System.out.println(agent1 + "-" + agent2);
		if (agent2 > agent1) {
			int temp = agent1;
			agent1 = agent2;
			agent2 = temp;
		}
		String agentPair = agent1 + "-" + agent2;
		/*System.out.println(history.get(new Agent[] {agents.get(1), agents.get(0)}));
		System.out.println(agents.get(agent1).id);*/
		//System.out.println(history.containsKey(agentPair));
		double[] historyPair = history.get(agentPair);
		/*if (print) {
			System.out.println("dialogs: " + historyPair[0]);
			System.out.println("successes: " + historyPair[1]);
		}*/
		if (historyPair[0] == 0) {
			/*if (print)
				System.out.println("weight: " + 0);*/
			return 0;
		}
		/*if (print)
			System.out.println("weight: " + historyPair[1] / historyPair[0]);*/
		return historyPair[1] / historyPair[0]; // Successes / dialogs
	}
	
	private static void incrementDialogs(int agent1, int agent2) {
		if (agent2 > agent1) {
			int temp = agent1;
			agent1 = agent2;
			agent2 = temp;
		}
		String agentPair = agent1 + "-" + agent2;
		double[] historyPair = history.get(agentPair);
		historyPair[0]++;
		history.put(agentPair, historyPair);
		//if (agent1 == 1 && agent2 == 0)
			//System.out.println(agent1 + "-" + agent2 + ": " + historyPair[0]);
	}
	
	private static void incrementSuccesses(int agent1, int agent2) {
		if (agent2 > agent1) {
			int temp = agent1;
			agent1 = agent2;
			agent2 = temp;
		}
		String agentPair = agent1 + "-" + agent2;
		double[] historyPair = history.get(agentPair);
		historyPair[1]++;
		history.put(agentPair, historyPair);
	}
	
	private static void addHistoryForChild(Agent child, int id, int parent1, int parent2) {
		for (int i = 0; i < agents.size(); i++) {
			
			double[] history1 = new double[] {0, 0};
			if (i != parent1) {
				if (i < parent1)
					history1 = history.get(parent1 + "-" + i);
				else
					history1 = history.get(i + "-" + parent1);
			}
			
			double[] history2 = new double[] {0, 0};
			if (i != parent2) {
				if (i < parent2)
					history2 = history.get(parent2 + "-" + i);
				else
					history2 = history.get(i + "-" + parent2);
			}
			
			double[] newHistory = new double[2];
			newHistory[0] = (history1[0] + history2[0]) / 2;
			newHistory[1] = (history1[1] + history2[1]) / 2;
			
			//System.out.println("c" + id + "-" + i + ": " + newHistory[0] + ", " + newHistory[1]);
			history.put("c" + id + "-" + i, newHistory);
		}
		
		for (int i = 0; i < children.size() - 1; i++) {
			
			double[] history1 = history.get("c" + i + "-" + parent1);
			double[] history2 = history.get("c" + i + "-" + parent2);
			
			double[] newHistory = new double[2];
			newHistory[0] = (history1[0] + history2[0]) / 2;
			newHistory[1] = (history1[1] + history2[1]) / 2;
			
			history.put("c" + id + "-c" + i, newHistory);
		}
	}
	
	private static void generationalHistoryCleanup() {
		Map<String, double[]> cleanHistory = new HashMap<>();
		for (Map.Entry<String, double[]> entry : history.entrySet()) {
			if (entry.getKey().startsWith("c") && entry.getKey().contains("-c")) {
				String[] ids = entry.getKey().split("-");
				String id = ids[0].substring(1) + "-" + ids[1].substring(1);
				cleanHistory.put(id, entry.getValue());
			}
		}
		history = cleanHistory;
	}
}
