package lipowski_lipowska;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Random;

import org.graphstream.graph.Graph;
import org.graphstream.graph.implementations.SingleGraph;

public class Main {
	
	private static List<Agent> agents;
	private static Random random = new Random();
	private static double[][] pairDialogs;
	private static double[][] pairSuccesses;
	private static int numberOfAgents;
	private static int numberOfGenerations;
	private static double epsilon;
	
	public static void main(String[] args) {
		numberOfGenerations = 10000;
		int[] noAgents = {100};
		for (int nAgents : noAgents) {
			//epsilon = 0.0001; // 10^-4
			epsilon = 0.00001; // 10^-5
			//epsilon = Math.sqrt(0.00001 / nAgents);
			numberOfAgents = nAgents;
			double[][] results = new double[10][numberOfGenerations];
			for (int i = 0; i < 10; i++) {
				results[i] = game();
				//System.out.println("GAME, gen1: " + results[i][0]);
			}
			//System.out.println(nAgents + " agents:");
			for (int i = 0; i < numberOfGenerations; i++) {
				double sum = 0;
				for (int j = 0; j < 10; j++) {
					sum += results[j][i];
				}
				//System.out.println(sum / 10.0);
			}
			
			/*Graph graph = new SingleGraph("Tutorial 1");
			for (int i = 0; i < nAgents; i++) {
				graph.addNode(Integer.toString(i));
			}
			for (int i = 0; i < nAgents; i++) {
				for (int j = 0; j < i; j++) {
					if (pairSuccesses[i][j] /pairDialogs[i][j] > 0.5)
						graph.addEdge(Integer.toString(i) + "-" + Integer.toString(j), Integer.toString(i), Integer.toString(j));
				}
			}
			graph.display();*/
		}
		
		
	}

	public static double[] game() {
		double[] results = new double[numberOfGenerations];
		
		agents = new ArrayList<>();
		for (int i = 0; i < numberOfAgents; i++) {
			agents.add(new Agent());
		}
		pairDialogs = new double[numberOfAgents][numberOfAgents];
		pairSuccesses = new double[numberOfAgents][numberOfAgents];
		for (int i = 0; i < pairDialogs.length; i++) {
			for (int j = 0; j < pairDialogs[i].length; j++) {
				pairDialogs[i][j] = 0;
				pairSuccesses[i][j] = 0; // Er denne loopen ndvendig?
			}
		}
		
		for (int i = 0; i < numberOfGenerations; i++) {
			/*if (i % 100 == 0)
				System.out.println("gen " + i);*/
			int numberOfSuccesses = 0;
			for (int j = 0; j < numberOfAgents; j++) {
				if (performDialog())
					numberOfSuccesses++;
			}
			Map<String, Integer> totalVocabulary = new HashMap<>();
			for (int j = 0; j < numberOfAgents; j++) {
				//System.out.print(agents.get(j).getVocabularySize());
				List<String> agentVocabulary = agents.get(j).getVocabulary();
				for (String word : agentVocabulary) {
					//System.out.println("ag" + j + ": " + word);
					if (totalVocabulary.containsKey(word))
						totalVocabulary.put(word, totalVocabulary.get(word) + 1);
					else
						totalVocabulary.put(word, 1);
				}
			}
			if (totalVocabulary.size() == 1) {
				for (int j = i; j < numberOfGenerations; j++) {
					results[j] = 1;
					//results[j] = 1 / numberOfAgents;
				}
				break;
			}
			//results[i] = 1.0 * Collections.max(totalVocabulary.values()) / numberOfAgents; // PRINT SHARE OF MOST COMMON LANGUAGE
			results[i] = totalVocabulary.size(); // PRINT NUMBER OF LANGUAGES
			//results[i] = 1.0 * totalVocabulary.size() / numberOfAgents; // PRINT RELATIVE NUMBER OF LANGUAGES
			//results[i] = 1.0 * numberOfSuccesses / numberOfAgents; // PRINT SUCCESS RATE
		}
		//System.out.println(numberOfSuccesses);
		//for (int i = 0; i < pairDialogs.length; i++) {
			//for (int j = 0; j < pairDialogs[i].length; j++) {
				//if (pairDialogs[i][j] > 0) {}
					//System.out.print(pairSuccesses[i][j] / pairDialogs [i][j]);
				//System.out.print(";");
			//}
			//System.out.println();
		//}
		return results;
	}
	
	private static boolean performDialog() {
		int speakerId = random.nextInt(numberOfAgents);
		Agent speaker = agents.get(speakerId);
		int hearerId = selectHearerId(speakerId);
		Agent hearer = agents.get(hearerId);
		//System.out.println("speaker " + speakerId + ", hearer " + hearerId);
		pairDialogs[speakerId][hearerId]++;
		pairDialogs[hearerId][speakerId]++;
		boolean success = dialog(speaker, hearer);
		if (success) {
			pairSuccesses[speakerId][hearerId]++;
			pairSuccesses[hearerId][speakerId]++;
		}
		
		return success;
	}
	
	private static int selectHearerId(int speakerId) {
		double totalWeights = 0;
		for (int i = 0; i < pairDialogs[speakerId].length; i++) {
			if (i == speakerId)
				continue;
			totalWeights += epsilon;
			if (pairDialogs[speakerId][i] > 0)
				totalWeights += pairSuccesses[speakerId][i] / pairDialogs[speakerId][i];
		}
		double randomNumber = totalWeights * random.nextDouble();
		double temp = 0;
		for (int i = 0; i < pairDialogs[speakerId].length; i++) {
			if (i == speakerId)
				continue;
			temp += epsilon;
			if (pairDialogs[speakerId][i] > 0)
				temp += pairSuccesses[speakerId][i] / pairDialogs[speakerId][i];
			if (temp > randomNumber)
				return i;
		}
		return -1;
	}
	
	private static boolean dialog(Agent speaker, Agent hearer) {
		String word = speaker.speak();
		boolean success = hearer.listen(word);
		if (success) {
			speaker.retainOnly(word);
			hearer.retainOnly(word);
		}
		else
			hearer.addWord(word);
		//System.out.print("DIALOG: " + word + " - ");
		//System.out.println(success ? "SUCCESS" : "FAILIURE");
		//System.out.println(success ? 1 : 0);
//		if (success)
//			System.out.println("success");
		return success;
	}
}
